import { Registry, Account, createBid } from './index';
import { createTestAccounts, getConfig } from './testing/helper';
import { DENOM } from './constants';
import { INVALID_BID_ERROR, OWNER_MISMATCH_ERROR, RELEASE_FUNDS_ERROR } from './types/cerc/auction/message';

jest.setTimeout(30 * 60 * 1000);
const { chainId, rpcEndpoint, gqlEndpoint, privateKey, fee } = getConfig();

const DURATION = 60;
const TX_FEES = Number(fee.amount[0].amount);
const COMMIT_FEE = '1000';
const REVEAL_FEE = '1000';

const CREATOR_INITIAL_BALANCE = 1000000000000;
const BIDDER_INITIAL_BALANCE = 20000000;
const LOWEST_BID_AMOUNT = 10000000;

const auctionTests = () => {
  let registry: Registry;
  let auctionId: string;

  let auctionCreatorAccount: { address: string, privateKey: string };
  let bidderAccounts: { address: string, privateKey: string, bid?: any }[] = [];

  const numBidders = 3;
  let bidAmounts: string[] = [];

  beforeAll(async () => {
    registry = new Registry(gqlEndpoint, rpcEndpoint, { chainId });

    // Create auction creator account
    const [auctionCreator] = await createTestAccounts(1);
    await registry.sendCoins({ denom: DENOM, amount: CREATOR_INITIAL_BALANCE.toString(), destinationAddress: auctionCreator.address }, privateKey, fee);
    auctionCreatorAccount = { address: auctionCreator.address, privateKey: auctionCreator.privateKey.toString('hex') };

    // Create bidder accounts
    const accounts = await createTestAccounts(numBidders);
    for (let i = 0; i < numBidders; i++) {
      await registry.sendCoins({ denom: DENOM, amount: BIDDER_INITIAL_BALANCE.toString(), destinationAddress: accounts[i].address }, privateKey, fee);
      bidderAccounts.push({ address: accounts[i].address, privateKey: accounts[i].privateKey.toString('hex') });
    }
  });

  test('Create a vickrey auction', async () => {
    const minimumBid = '1000000';

    const auction = await registry.createAuction(
      {
        commitsDuration: DURATION.toString(),
        revealsDuration: DURATION.toString(),
        denom: DENOM,
        commitFee: COMMIT_FEE,
        revealFee: REVEAL_FEE,
        minimumBid
      },
      auctionCreatorAccount.privateKey,
      fee
    );

    expect(auction.auction?.id).toBeDefined();

    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    auctionId = auction.auction!.id;
    expect(auction.auction?.status).toEqual('commit');
  });

  test('Commit bids.', async () => {
    for (let i = 0; i < numBidders; i++) {
      bidAmounts.push((LOWEST_BID_AMOUNT + (i * 500)).toString());
      bidderAccounts[i].bid = await createBid(chainId, auctionId, bidderAccounts[i].address, `${bidAmounts[i]}${DENOM}`);
      await registry.commitBid({ auctionId, commitHash: bidderAccounts[i].bid.commitHash }, bidderAccounts[i].privateKey, fee);
    }

    const [auction] = await registry.getAuctionsByIds([auctionId]);
    expect(auction.status).toEqual('commit');
    expect(auction.bids.length).toEqual(3);
    auction.bids.forEach((bid: any) => {
      expect(bid.status).toEqual('commit');
    });
  });

  test('Wait for reveal phase.', (done) => {
    const commitTime = DURATION * 1000;
    const waitTime = commitTime + (6 * 1000);

    setTimeout(done, waitTime);
  });

  test('Reveal bids.', async () => {
    const [auction] = await registry.getAuctionsByIds([auctionId]);
    expect(auction.status).toEqual('reveal');

    for (let i = 0; i < numBidders; i++) {
      await registry.revealBid({ auctionId, reveal: bidderAccounts[i].bid.revealString }, bidderAccounts[i].privateKey, fee);
    }
  });

  test('Check bids are revealed', async () => {
    const [auction] = await registry.getAuctionsByIds([auctionId]);
    expect(auction.status).toEqual('reveal');
    auction.bids.forEach((bid: any) => {
      expect(bid.status).toEqual('reveal');
    });

    const expectedBidAmounts = bidAmounts.map(bidAmount => { return { quantity: bidAmount.toString(), type: DENOM }; });
    const actualBidAmounts = auction.bids.map((bid: any) => bid.bidAmount);
    expect(actualBidAmounts).toEqual(expect.arrayContaining(expectedBidAmounts));

    // Check that the bid amounts are locked after reveal phase
    for (let i = 0; i < numBidders; i++) {
      // The bid amount, commit fee, reveal fee and tx fees (for commit and reveal txs) will be deducted from the bidder's acoount
      const expectedBalance = BIDDER_INITIAL_BALANCE - parseInt(bidAmounts[i]) - (2 * TX_FEES) - parseInt(COMMIT_FEE) - parseInt(REVEAL_FEE);
      await checkAccountBalance(registry, bidderAccounts[i].address, expectedBalance);
    }
  });

  test('Wait for auction completion.', (done) => {
    const revealTime = DURATION * 1000;
    const waitTime = revealTime + (6 * 1000);

    setTimeout(done, waitTime);
  });

  test('Check auction status, winner address and balance.', async () => {
    const [auction] = await registry.getAuctionsByIds([auctionId]);
    expect(auction.status).toEqual('completed');

    const highestBidder = bidderAccounts[bidderAccounts.length - 1];
    const secondHighestBidder = bidderAccounts[bidderAccounts.length - 2];

    expect(auction.winnerAddresses).toHaveLength(1);

    expect(auction.winnerAddresses[0]).toEqual(highestBidder.address);
    expect(highestBidder.bid.reveal.bidAmount).toEqual(`${auction.winnerBids[0].quantity}${auction.winnerBids[0].type}`);
    expect(secondHighestBidder.bid.reveal.bidAmount).toEqual(`${auction.winnerPrice.quantity}${auction.winnerPrice.type}`);

    const winningPriceAmount = parseInt(auction.winnerPrice.quantity);

    // The winning price will get deducted after auction completion
    const expectedBalance = BIDDER_INITIAL_BALANCE - winningPriceAmount - (2 * TX_FEES) - parseInt(COMMIT_FEE);
    await checkAccountBalance(registry, highestBidder.address, expectedBalance);
  });
};

const providerAuctionTestsWithBids = (bidAmounts: number[], numProviders: number) => {
  let registry: Registry;
  let auctionId: string;

  let auctionCreatorAccount: { address: string, privateKey: string };
  let bidderAccounts: { address: string, privateKey: string, bid?: any }[] = [];
  let sortedBidders: { address: string, privateKey: string, bid?: any }[] = [];
  let filteredBidders: { address: string, privateKey: string, bid?: any }[] = [];
  let invalidBidderAddress: string;
  let otherAccount: { address: string, privateKey: string };

  const numBidders = bidAmounts.length;
  const maxPrice = 10 * LOWEST_BID_AMOUNT;

  beforeAll(async () => {
    registry = new Registry(gqlEndpoint, rpcEndpoint, { chainId });

    // Create auction creator account
    const [auctionCreator] = await createTestAccounts(1);
    await registry.sendCoins({ denom: DENOM, amount: CREATOR_INITIAL_BALANCE.toString(), destinationAddress: auctionCreator.address }, privateKey, fee);
    auctionCreatorAccount = { address: auctionCreator.address, privateKey: auctionCreator.privateKey.toString('hex') };

    // Create bidder accounts
    const accounts = await createTestAccounts(numBidders);
    for (let i = 0; i < numBidders; i++) {
      await registry.sendCoins({ denom: DENOM, amount: BIDDER_INITIAL_BALANCE.toString(), destinationAddress: accounts[i].address }, privateKey, fee);
      bidderAccounts.push({ address: accounts[i].address, privateKey: accounts[i].privateKey.toString('hex') });
    }

    const mnenonic3 = Account.generateMnemonic();
    const otherAcc = await Account.generateFromMnemonic(mnenonic3);
    await otherAcc.init();

    await registry.sendCoins({ denom: DENOM, amount: CREATOR_INITIAL_BALANCE.toString(), destinationAddress: otherAcc.address }, privateKey, fee);
    otherAccount = { address: otherAcc.address, privateKey: otherAcc.privateKey.toString('hex') };
  });

  test('Create a provider auction', async () => {
    const auction = await registry.createProviderAuction(
      {
        commitsDuration: DURATION.toString(),
        revealsDuration: DURATION.toString(),
        denom: DENOM,
        commitFee: COMMIT_FEE,
        revealFee: REVEAL_FEE,
        maxPrice: maxPrice.toString(),
        numProviders
      },
      auctionCreatorAccount.privateKey, fee
    );

    expect(auction.auction?.id).toBeDefined();

    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    auctionId = auction.auction!.id;
    expect(auction.auction?.status).toEqual('commit');

    // The locked amount and tx fees are deducted from the auction creator's account
    const expectedBalance = CREATOR_INITIAL_BALANCE - (maxPrice * numProviders) - TX_FEES;
    await checkAccountBalance(registry, auctionCreatorAccount.address, expectedBalance);
  });

  test('Commit bids.', async () => {
    for (let i = 0; i < numBidders; i++) {
      bidderAccounts[i].bid = await createBid(chainId, auctionId, bidderAccounts[i].address, `${bidAmounts[i]}${DENOM}`);
      await registry.commitBid({ auctionId, commitHash: bidderAccounts[i].bid.commitHash }, bidderAccounts[i].privateKey, fee);
    }

    sortedBidders = bidderAccounts.slice().sort((a, b) => {
      return parseInt(a.bid.reveal.bidAmount) - parseInt(b.bid.reveal.bidAmount);
    });

    for (const bidder of sortedBidders) {
      if (parseInt(bidder.bid.reveal.bidAmount) > maxPrice) {
        invalidBidderAddress = bidder.address;
      }
    }

    const [auction] = await registry.getAuctionsByIds([auctionId]);
    expect(auction.status).toEqual('commit');
    expect(auction.bids.length).toEqual(numBidders);
    auction.bids.forEach((bid: any) => {
      expect(bid.status).toEqual('commit');
    });
  });

  test('Wait for reveal phase.', (done) => {
    const commitTime = DURATION * 1000;
    const waitTime = commitTime + (6 * 1000);

    setTimeout(done, waitTime);
  });

  test('Reveal bids.', async () => {
    const [auction] = await registry.getAuctionsByIds([auctionId]);
    expect(auction.status).toEqual('reveal');

    for (let i = 0; i < numBidders; i++) {
      try {
        await registry.revealBid(
          { auctionId, reveal: bidderAccounts[i].bid.revealString },
          bidderAccounts[i].privateKey,
          fee
        );
      } catch (error: any) {
        if (invalidBidderAddress === bidderAccounts[i].address) {
          expect(error.toString()).toContain(INVALID_BID_ERROR);
        } else {
          throw error;
        }
      }
    }
  });

  test('Check bids are revealed', async () => {
    const [auction] = await registry.getAuctionsByIds([auctionId]);
    expect(auction.status).toEqual('reveal');

    const expectedBids = sortedBidders.map((bidder: any) => {
      if (invalidBidderAddress === bidder.address) {
        return '0';
      }
      return bidder.bid.reveal.bidAmount;
    });

    const actualBidAmounts = auction.bids.map((bid: any) => `${bid.bidAmount.quantity}${bid.bidAmount.type}`);
    expect(actualBidAmounts).toEqual(expect.arrayContaining(expectedBids));
  });

  test('Wait for auction completion.', (done) => {
    const revealTime = DURATION * 1000;
    const waitTime = revealTime + (6 * 1000);

    setTimeout(done, waitTime);
  });

  test('Check auction status, winners and creator balance.', async () => {
    const [auction] = await registry.getAuctionsByIds([auctionId]);
    expect(auction.status).toEqual('completed');

    // Filter bidders to exclude bids exceeding maxPrice
    filteredBidders = sortedBidders.filter(bidder => {
      return parseInt(bidder.bid.reveal.bidAmount) <= parseInt(`${maxPrice}${DENOM}`);
    });

    const numWinners = Math.min(filteredBidders.length, numProviders);
    const winningBidders = filteredBidders.slice(0, numWinners);

    // Check winner price is equal to highest winning bid
    const expectedWinningPrice = winningBidders[numWinners - 1].bid.reveal.bidAmount;
    expect(`${auction.winnerPrice.quantity}${auction.winnerPrice.type}`).toEqual(expectedWinningPrice);

    // The total winning amount will get deducted and the leftover locked amount is returned
    const totalWinningAmount = parseInt(auction.winnerPrice.quantity) * winningBidders.length;
    const expectedCreatorBalance = CREATOR_INITIAL_BALANCE - totalWinningAmount - TX_FEES;
    await checkAccountBalance(registry, auctionCreatorAccount.address, expectedCreatorBalance);

    // Funds should not be given to winners on auction completion
    for (let i = 0; i < winningBidders.length; i++) {
      const bidWinner = winningBidders[i];

      expect(auction.winnerAddresses[i]).toEqual(bidWinner.address);
      expect(`${auction.winnerBids[i].quantity}${auction.winnerBids[i].type}`).toEqual(bidWinner.bid.reveal.bidAmount);

      const expectedBidderBalance = BIDDER_INITIAL_BALANCE - (2 * TX_FEES) - parseInt(COMMIT_FEE);
      await checkAccountBalance(registry, bidWinner.address, expectedBidderBalance);
    }
  });

  test('Release funds and check provider balances.', async () => {
    const [auction] = await registry.getAuctionsByIds([auctionId]);

    const numWinners = Math.min(filteredBidders.length, numProviders);
    const winningBidders = filteredBidders.slice(0, numWinners);
    const losingBidders = bidderAccounts.filter(bidder => !winningBidders.includes(bidder));

    const winningBidAmount = parseInt(auction.winnerPrice.quantity);

    // Only owner can release funds
    try {
      await registry.releaseFunds({ auctionId }, otherAccount.privateKey, fee);
    } catch (error: any) {
      expect(error.toString()).toContain(OWNER_MISMATCH_ERROR);
    }

    // Release funds
    const auctionObj = await registry.releaseFunds({ auctionId }, auctionCreatorAccount.privateKey, fee);
    expect(auctionObj.auction?.fundsReleased).toBe(true);

    // Auction winners get the winning amount after funds are released
    const expectedBidderBalance = BIDDER_INITIAL_BALANCE + winningBidAmount - (2 * TX_FEES) - parseInt(COMMIT_FEE);
    for (let i = 0; i < winningBidders.length; i++) {
      const bidWinner = winningBidders[i];

      expect(auction.winnerAddresses[i]).toEqual(bidWinner.address);
      expect(`${auction.winnerBids[i].quantity}${auction.winnerBids[i].type}`).toEqual(bidWinner.bid.reveal.bidAmount);

      await checkAccountBalance(registry, bidWinner.address, expectedBidderBalance);
    }

    // Check balances of non-winners
    for (const bidder of losingBidders) {
      let expectedBidderBalance: number;
      if (invalidBidderAddress !== bidder.address) {
        expectedBidderBalance = BIDDER_INITIAL_BALANCE - (2 * TX_FEES) - parseInt(COMMIT_FEE);
      } else {
        // Bid is invalid, reveal fees are not returned
        expectedBidderBalance = BIDDER_INITIAL_BALANCE - (2 * TX_FEES) - parseInt(COMMIT_FEE) - parseInt(REVEAL_FEE);
      }

      await checkAccountBalance(registry, bidder.address, expectedBidderBalance);
    }

    // Funds cannot be released twice
    try {
      await registry.releaseFunds({ auctionId }, auctionCreatorAccount.privateKey, fee);
    } catch (error: any) {
      expect(error.toString()).toContain(RELEASE_FUNDS_ERROR);
    }
  });
};

const checkAccountBalance = async (registry: Registry, address: string, expectedBalance: number): Promise<void> => {
  const [winnerAccountObj] = await registry.getAccounts([address]);
  const [{ type, quantity }] = winnerAccountObj.balance;
  const actualBalance = parseInt(quantity);

  expect(actualBalance).toBe(expectedBalance);
  expect(type).toBe(DENOM);
};

const providerAuctionTests = () => {
  // With a bid amount greater than the max price
  const bids = [10002000, 10009000, 10006000 * 10, 10004000];

  // Number of providers is less than number of bidders
  describe('Auction with numProviders < numBidders', () => providerAuctionTestsWithBids(bids, 3));
  // Number of providers is greater than number of bidders
  describe('Auction numProviders > numBidders', () => providerAuctionTestsWithBids(bids, 5));
};

describe('Vickrey Auction', () => auctionTests());
describe('Provider Auction', () => providerAuctionTests());
